-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all; 
use work.sha3_pkg.all;
use work.sha3_blake_package.all;

entity blake_control_ppl is		
	generic ( 	 	  
		version : integer := SHA3_ROUND3;
		FF : integer := 2;	
		ppl : integer := 4;
		IW : integer := 32;
		BS : integer := 512;
		HS : integer := 256;
		W : integer := 64
	);
	port (					
		rst			: in std_logic;
		clk			: in std_logic;
		
		-- datapath signals
		--fsm1		   
		extra_block 			: in std_logic_vector(ppl-1 downto 0);				
		zcRemainder, zc0, zc1	: in std_logic_vector(ppl-1 downto 0);		
		final_segment 			: in std_logic_vector(ppl-1 downto 0);		
		
		ein, ec, lc, stbot		: out std_logic_vector(ppl-1 downto 0);				
		dt, eth, etl			: out std_logic_vector(ppl-1 downto 0);
		lm 						: out std_logic_vector(ppl-1 downto 0);
		load_id 				: out std_logic_vector(ppl-1 downto 0);
		
			--fsm2
		smsg : out std_logic_vector(log2(ppl)-1 downto 0);		
		load_h : out std_logic;
		sf : out std_logic;
		round 		: out std_logic_vector(3+FF-1 downto 0);
		eh, load_hash : out std_logic_vector(ppl-1 downto 0);
		sperm : out std_logic;
		-- FSM3
		shift_dout		: out std_logic_vector(ppl-1 downto 0);
		sout 		: out std_logic_vector(log2(ppl)-1 downto 0);
		sid : out std_logic;
			--top fsm			
		
		-- fifo signals
		src_ready	: in std_logic_vector(ppl-1 downto 0);
		src_read	: out std_logic_vector(ppl-1 downto 0);
		dst_ready 	: in std_logic;
		dst_write	: out std_logic
	);				 
end blake_control_ppl;

architecture struct of blake_control_ppl is				   	 
	-- fsm1
	signal block_ready_set, msg_end_set, load_next_block : std_logic_vector(ppl-1 downto 0);	
	-- fsm2						
		-- fsm1 communications		
	signal block_ready_clr, msg_end_clr : std_logic_vector(ppl-1 downto 0); --out
	signal block_ready, msg_end : std_logic_vector(ppl-1 downto 0); --in
		-- fsm2 communications
	signal output_busy_set : std_logic_vector(ppl-1 downto 0); --out
	signal output_busy : std_logic_vector(ppl-1 downto 0); --in 
	
	-- fsm3										
	signal output_busy_clr : std_logic_vector(ppl-1 downto 0); 	
	signal shift_dout_fsm3, load_hash_fsm2 : std_logic_vector(ppl-1 downto 0);
	
	-- pipelined signals
	signal er_sig, sf_sig, dst_write_sig, load_h_sig  : std_logic;
	signal sid_sig, lm_fsm2, sperm_sig : std_logic;	
	signal eh_sig : std_logic_vector(ppl-1 downto 0);
	signal sout_sig, smsg_sig : std_logic_vector(log2(ppl)-1 downto 0);
	signal round_sig : std_logic_vector(3+FF-1 downto 0);
begin
	
	gfsm1: for i in 0 to ppl-1 generate
		fsm1_gen : entity work.blake_fsm1(counter)
			generic map (BS => BS)
			port map (
			clk => clk, rst => rst, 
			final_segment => final_segment(i), zcRemainder => zcRemainder(i), zc0 => zc0(i), zc1 => zc1(i),
				ein => ein(i), ec => ec(i), lc => lc(i), eth => eth(i), dt => dt(i), etl => etl(i),
				extra_block => extra_block(i), stbot => stbot(i), 
			load_next_block => load_next_block(i), block_ready_set => block_ready_set(i), msg_end_set => msg_end_set(i),
			src_ready => src_ready(i), src_read => src_read(i), load_id => load_id(i)
		);	 
	end generate;
	
	fsm2_gen : entity work.blake_fsm2(beh) 
		generic map ( ppl => ppl, h => HS, FF => FF, version => version)
		port map (
		clk => clk, rst => rst, 
		eh => eh_sig, smsg => smsg_sig, load_h => load_h_sig, sf => sf_sig, round => round_sig, load_hash => load_hash_fsm2, lm => lm_fsm2, sperm => sperm_sig,
		block_ready_clr => block_ready_clr, msg_end_clr => msg_end_clr, block_ready => block_ready, msg_end => msg_end,
		output_busy_set => output_busy_set, output_busy => output_busy
	); 
	
	fsm3_gen : entity work.sha3_fsm3_ppl(beh)
		generic map ( h => hs, w=> w, ppl => ppl )
		port map (
		clk => clk, rst => rst, 
		shift_dout => shift_dout_fsm3, sout => sout_sig,  sid => sid_sig,
		output_busy => output_busy, output_busy_clr => output_busy_clr,
		dst_ready => dst_ready, dst_write => dst_write_sig
	);	 	 		  
	
	outputreg: process( clk ) 
	begin
		if rising_edge( clk ) then
			eh <= eh_sig;
			smsg <= smsg_sig;
			load_h <= load_h_sig;
			sid <= sid_sig;		
			sf <= sf_sig;
			round <= round_sig;
			sperm <= sperm_sig;
			sout <= sout_sig;
			shift_dout <= shift_dout_fsm3 or load_hash_fsm2;
			load_hash <= load_hash_fsm2;	 
			dst_write <= dst_write_sig;	
		end if;
	end process;
		
	comm :  for i in 0 to ppl-1 generate
		lm(i) <= lm_fsm2;
		load_next_block(i) <= not block_ready(i);
		sr_blk_ready : sr_reg 
			port map ( rst => rst, clk => clk, set => block_ready_set(i), clr => block_ready_clr(i), output => block_ready(i));	
		sr_msg_end : sr_reg 
			port map ( rst => rst, clk => clk, set => msg_end_set(i), clr => msg_end_clr(i), output => msg_end(i));	
		sr_output_write : sr_reg 
			port map ( rst => rst, clk => clk, set => output_busy_set(i), clr => output_busy_clr(i), output => output_busy(i));			
	end generate; 
end struct;