-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.sha3_blake_package.ALL;
use work.sha3_pkg.all;

entity blake_datapath_ppl is	 	
	generic (	  	   
		FF : integer := 4;	
		PPL : integer := 4;
		BS : integer := 512;
		W : integer := 64;
		IW : integer := 32;
		HS : integer := 256
		);
	port (
		-- external
		rst 							: in std_logic;
		clk							 	: in std_logic;
		din  							: in std_logic_vector(ppl*w-1 downto 0);	
		dout 							: out std_logic_vector(w-1 downto 0);	
		--fsm 1		
		ein 							: in std_logic_vector(ppl-1 downto 0);
		ec, lc, lm, stbot 				: in std_logic_vector(ppl-1 downto 0);
		zcRemainder, zc0, zc1 			: out std_logic_vector(ppl-1 downto 0);
		extra_block 					: out std_logic_vector(ppl-1 downto 0);
		dt, eth, etl 					: in std_logic_vector(ppl-1 downto 0);		
		--fsm 2	
		load_h 							: in std_logic;
		eh 								: in std_logic_vector(ppl-1 downto 0);
		smsg 							: in std_logic_vector(log2(ppl)-1 downto 0);
		round 							: in std_logic_vector(3+FF-1 downto 0);				
		sf 								: in std_logic;
		sperm 							: in std_logic;
		
		-- Output				
		shift_dout, load_id, load_hash 	: in std_logic_vector(ppl-1 downto 0);
		sout 							: in std_logic_vector(log2(ppl)-1 downto 0);
		sid 							: in std_logic
	);				  
end blake_datapath_ppl;

architecture struct of blake_datapath_ppl is   
	
	-- CONSTANTS
	constant mw 			: integer := BS;		 
	constant bseg			: integer := BS/w;		  
	constant log2b			: integer := log2( BS );
	constant log2bzeros		: std_logic_vector(log2b-1 downto 0) := (others => '0');	
	constant bzeros			: std_logic_vector(BS-1 downto 0) := (others => '0');
	constant bhalfzeros		: std_logic_vector(BS/2-1 downto 0) := (others => '0');
	constant bmin	: integer := BS - 2 - iw*2;
	constant iv : std_logic_vector(BS/2-1 downto 0) := get_iv( HS, IW );	
	type bot_permute_type is array ( 0 to 15 ) of integer;
	constant bot_permute : bot_permute_type := ( 0,1,2,3,5,6,7,4,10,11,8,9,15,12,13,14 );	
	constant blake_constant : std_logic_vector(BS-1 downto 0) := get_cons( HS, BS, iw );
	-------------
	
	-- //////////////////////
	-- FUNCTIONS 
	type std_logic_matrix is array (natural range <>) of std_logic_vector(iw - 1 downto 0) ;
	function wordmatrix2blk  	(x : std_logic_matrix) return std_logic_vector is
		variable retval : std_logic_vector(BS-1 downto 0);
	begin
		for i in 0 to 15 loop
			retval(iw*(i+1) - 1 downto iw*i) := x(15-i);
		end loop;
		return retval;
	end wordmatrix2blk;
	function blk2wordmatrix  	(x : std_logic_vector; blksize : integer ) return std_logic_matrix is
		variable retval : std_logic_matrix(0 to blksize-1);
	begin
		for i in 0 to blksize-1 loop
			retval(blksize-1-i) := x(iw*(i+1) - 1 downto iw*i);
		end loop;
		return retval;
	end blk2wordmatrix;	
	-------------
	
	-- //////////////////////
	-- type declaration
	type c_array_type is array ( 0 to ppl-1) of std_logic_vector(15 downto 0);	
	type block_array_type is array (0 to ppl-1) of std_logic_vector(BS-1 downto 0);
	type half_block_array_type is array (0 to ppl-1) of std_logic_vector(BS/2-1 downto 0);	
	type t_top_type is array ( 0 to ppl-1 ) of std_logic_vector(2*iw-log2b-1 downto 0);
	type t_bot_type is array ( 0 to ppl-1 ) of std_logic_vector(log2b-1 downto 0);
	type t_type is array ( 0 to ppl-1 ) of std_logic_vector(2*iw-1 downto 0);
	-------------

	signal v1, v2, v2_perm, v2_revert, v2_regged, v3 : std_logic_matrix( 0 to 15 ); 
	signal cp : std_logic_matrix(0 to 16/FF - 1);	 
	
	-- FSM 1 (PROTOCOL / LOADING )
	signal c : c_array_type;
	signal t_top : t_top_type;
	signal t_bot, t_bot_enabled : t_bot_type;
	signal t : t_type;
	signal min, m : block_array_type;
	signal t_selected : std_logic_vector(2*iw-1 downto 0);
	signal m_selected : std_logic_vector(BS-1 downto 0);
	
	-- FSM2 ( ROUND )
	signal consout : std_logic_vector(BS/FF-1 downto 0);  			  -- 8 = b size, 4 = b/2 size	
	signal rdprime, rmux,chain_selected : std_logic_vector(BS/2 - 1 downto 0);
	signal rinit, rin, r, rprime, rprime_regged : std_logic_vector(BS-1 downto 0);
	signal chain : half_block_array_type;
	
	signal t_bot0 : std_logic_vector(ppl-1 downto 0);
	-- FSM 3 ( PROTOCOL / OUTPUT )
	-- id, hash		
	constant id_size : integer := 16;
	signal id_in : std_logic_vector(ppl*id_size-1 downto 0);
	
	-- debug			  
	-- signal rinit_m, r_m, cp_m : std_logic_matrix(0 to 15);				
begin		
	-- debug
	-- rinit_m <= blk2wordmatrix( rin, 16 );		 
	-- r_m <= blk2wordmatrix( r, 16 );	
	--		

	-- ///////////////////////////////////////////////////////
	-- FSM1 SECTION
	fsm1_section : for i in 0 to ppl-1 generate
		decounter_gen : decountern generic map ( N => 16, sub => 1 ) 
			port map ( 	clk => clk, rst => '0', load => lc(i), en => ec(i), 
						input => din(log2b+15+(ppl*w-(i+1)*w) downto log2b+(ppl*w-(i+1)*w)), output => c(i) );
		zc0(i) <= '1' when c(i) = 0  else '0'; 	
		zc1(i) <= '1' when c(i) = 1  else '0';
		
		th_gen : countern generic map ( N => 2*iw-log2b, step => 1 ) 
			port map ( 	clk => clk, rst => dt(i), load => '0', en => eth(i), 
						input => conv_std_logic_vector(0,2*iw-log2b), output => t_top(i) );
		tl_gen : regn generic map ( N => log2b, init => log2bzeros ) 
			port map ( 	clk => clk, rst => dt(i), en => etl(i), 
						input => din(log2b-1+ppl*w-(i+1)*w downto ppl*w-(i+1)*w), output => t_bot(i) );
		t_bot_enabled(i) <= t_bot(i) when (stbot(i) = '1') else (others => '0');
		t(i) <= t_top(i) & t_bot_enabled(i);	
		
		t_bot0(i) <= '0' when (t_bot(i) = 0) else '1';
		zcRemainder(i) <= '1' when (t_bot0(i) = '1') else '0';
		extra_block(i) <= '1' when ((t_bot(i) > BS-(IW*2+2)) or (t_bot0(i) = '0')) else '0';
		shfin_gen : sipo generic map ( N => BS,M =>w) 
			port map (	clk => clk, en => ein(i), 
						input => din(ppl*w-i*w-1 downto ppl*w-(i+1)*w), output => min(i) );
		mreg : regn generic map (N => mw, init => bzeros) 
			port map (clk => clk, rst => '0', en => lm(i), input => min(i), output => m(i) );
	end generate;

	
		
	-- ///////////////////////////////////////////////////////
	-- FSM2 SECTION
	ppl2_sel : if ppl = 2 generate
		t_selected <= t(0) when smsg(0) = '0' else t(1);
		m_selected <= m(0) when smsg(0) = '0' else m(1);
		chain_selected <= chain(0) when smsg(0) = '0' else chain(1);
	end generate;  
	ppl4_sel : if ppl = 4 generate
		with smsg(1 downto 0) select
		t_selected 	<= t(0) when "00", t(1) when "01", t(2) when "10",t(3) when others;	
		with smsg(1 downto 0) select
		m_selected 	<= m(0) when "00", m(1) when "01", m(2) when "10",m(3) when others;	
		with smsg(1 downto 0) select
		chain_selected 	<= chain(0) when "00", chain(1) when "01", chain(2) when "10",chain(3) when others;
	end generate;
	
	rmux <= iv when sf = '1' else rdprime;

	rinit <= rmux & blake_constant(BS-1 downto BS-iw*4)  & 
			(t_selected(iw-1 downto 0) xor blake_constant(BS-1-iw*4 downto BS-iw*5)) &  (t_selected(iw-1 downto 0) xor blake_constant(BS-iw*5-1 downto BS-iw*6)) &
			(t_selected(2*iw-1 downto iw) xor blake_constant(BS-1-iw*6 downto BS-iw*7))  & (t_selected(2*iw-1 downto iw) xor blake_constant(BS-iw*7-1 downto BS-iw*8));
	
	fsm2_section : for i in 0 to ppl-1 generate
		chaining_value_gen : regn generic map ( N => BS/2, init => bhalfzeros ) 
			port map (clk => clk, rst => '0', en => eh(i), input => rmux, output => chain(i) );
	end generate;
	
	rin <= rinit when (load_h = '1') else rprime;
	r_gen : regn generic map ( N => BS, init => bzeros ) 
		port map (clk => clk, rst => '0', en => '1', input => rin, output => r );
	
	-- //////////////
	-- CORE OPERATION	
	g_gfunc4 : if FF = 2 generate
		v1 <= blk2wordmatrix(r, 16);	
		perm4_gen : entity work.permute4xor_ppl(muxbased) 
			generic map (ppl => ppl, h => HS, b => BS, iw => iw) 
			port map ( clk => clk, m => m_selected, round => round, consout => consout );
		cp <= blk2wordmatrix(consout,8);	
		glvl1 : for i in 0 to 3 generate		   
			ppl2 : if ppl = 2 generate
				g0123 : entity work.gfunc_modified_ppl2(struct) 	generic map ( iw => iw, h => HS )
							port map (clk => clk,
									ain => v1(i),bin => v1(i+4),cin => v1(i+8),din => v1(i+12),
									const_0 => cp(2*i),const_1 => cp(2*i + 1),
									aout =>  v2(i), bout => v2(i+4), cout=> v2(i+8),dout => v2(i+12));								
			end generate;
			ppl4 : if ppl = 4 generate 
				g0123 : entity work.gfunc_modified_ppl4(struct) 	generic map ( iw => iw, h => HS )
							port map (clk => clk,
									ain => v1(i),bin => v1(i+4),cin => v1(i+8),din => v1(i+12),
									const_0 => cp(2*i),const_1 => cp(2*i + 1),
									aout =>  v2(i), bout => v2(i+4), cout=> v2(i+8),dout => v2(i+12));								
			end generate;
		end generate;	
		v2_gen : for i in 0 to 15 generate
			v2_perm( bot_permute(i) ) <= v2( i ); 
			v2_revert( i ) <= v2( bot_permute(i) );
		end generate;	
		v3 <= v2_perm when sperm = '0' else v2_revert;		 
	end generate;	
	
	g_gfunc8 : if FF = 1 generate
		perm8_gen : entity work.permute8xor_ppl(muxbased) 
			generic map (ppl => ppl, h => HS, b => BS, iw => iw) 
			port map ( clk => clk,  m => m_selected, round => round, consout => consout );
				
		v1 <= blk2wordmatrix(r, 16);
		cp <= blk2wordmatrix(consout, 16);
		
		v2reg : process( clk ) 
		begin	
			if rising_edge( clk ) then
				v2_regged <= v2;
			end if;
		end process;
			
		ppl4: if ppl = 4 generate
			glvl1 : for i in 0 to 3 generate
				g0123 : entity work.gfunc8x_modified_ppl2(struct) 	generic map ( iw => iw, h => HS )
					port map ( 		clk => clk,
									ain => v1(i),bin => v1(i+4),cin => v1(i+8),din => v1(i+12),
									const_0 => cp(2*i),const_1 => cp(2*i + 1),
									aout =>  v2(i), bout => v2(i+4), cout=> v2(i+8),dout => v2(i+12));
			end generate;
							
			
			
			g4 : entity work.gfunc8x_modified_ppl2(struct) generic map ( iw => iw, h => HS )
					port map ( 	clk => clk,
								ain => v2_regged(0),bin => v2_regged(5),cin => v2_regged(10),din => v2_regged(15),
								const_0 => cp(8),const_1 => cp(9),
								aout =>  v3(0), bout => v3(5), cout=> v3(10),dout => v3(15));
															 
			g5 : entity work.gfunc8x_modified_ppl2(struct) generic map ( iw => iw, h => HS )
				port map ( 	clk => clk,
								ain => v2_regged(1),bin => v2_regged(6),cin => v2_regged(11),din => v2_regged(12),
								 const_0 => cp(10),const_1 => cp(11),
								  aout =>  v3(1), bout => v3(6), cout=> v3(11),dout => v3(12));
		
			g6 : entity work.gfunc8x_modified_ppl2(struct) generic map ( iw => iw, h => HS )
				port map ( 	clk => clk,
									ain => v2_regged(2),bin => v2_regged(7),cin => v2_regged(8),din => v2_regged(13),
								  const_0 => cp(12),const_1 => cp(13),
								  aout =>  v3(2), bout => v3(7), cout=> v3(8),dout => v3(13));
		
			g7 : entity work.gfunc8x_modified_ppl2(struct) generic map ( iw => iw, h => HS )
				port map ( 	clk => clk,
								ain => v2_regged(3),bin => v2_regged(4),cin => v2_regged(9),din => v2_regged(14),
								  const_0 => cp(14),const_1 => cp(15),
								  aout =>  v3(3), bout => v3(4), cout=> v3(9),dout => v3(14));
		end generate;	
		
		ppl2: if ppl = 2 generate
			glvl1 : for i in 0 to 3 generate
				g0123 : entity work.gfunc_modified(struct) 	generic map ( iw => iw, h => HS )
								port map ( ain => v1(i),bin => v1(i+4),cin => v1(i+8),din => v1(i+12),
									  const_0 => cp(2*i),const_1 => cp(2*i + 1),
									  aout =>  v2(i), bout => v2(i+4), cout=> v2(i+8),dout => v2(i+12));
			end generate;
							
			
			
			g4 : entity work.gfunc_modified(struct) generic map ( iw => iw, h => HS )
					   port map ( ain => v2_regged(0),bin => v2_regged(5),cin => v2_regged(10),din => v2_regged(15),
								  const_0 => cp(8),const_1 => cp(9),
								  aout =>  v3(0), bout => v3(5), cout=> v3(10),dout => v3(15));
															 
			g5 : entity work.gfunc_modified(struct) generic map ( iw => iw, h => HS )
					   port map ( ain => v2_regged(1),bin => v2_regged(6),cin => v2_regged(11),din => v2_regged(12),
								 const_0 => cp(10),const_1 => cp(11),
								  aout =>  v3(1), bout => v3(6), cout=> v3(11),dout => v3(12));
		
			g6 : entity work.gfunc_modified(struct) generic map ( iw => iw, h => HS )
			 		   port map ( ain => v2_regged(2),bin => v2_regged(7),cin => v2_regged(8),din => v2_regged(13),
								  const_0 => cp(12),const_1 => cp(13),
								  aout =>  v3(2), bout => v3(7), cout=> v3(8),dout => v3(13));
		
			g7 : entity work.gfunc_modified(struct) generic map ( iw => iw, h => HS )
					   port map ( ain => v2_regged(3),bin => v2_regged(4),cin => v2_regged(9),din => v2_regged(14),
								  const_0 => cp(14),const_1 => cp(15),
								  aout =>  v3(3), bout => v3(4), cout=> v3(9),dout => v3(14));
		end generate;
		
								  
		rprime <= wordmatrix2blk(v3);
	end generate;
	
	rprime <= wordmatrix2blk(v3);		
	
	rpr : process ( clk ) 
	begin	
		if rising_edge( clk ) then
			rprime_regged <= rprime;
		end if;
	end process;
		
	--finalization				
	rdprime <= chain_selected xor rprime_regged(BS-1 downto BS/2) xor rprime_regged(BS/2-1 downto 0);
	
	-- ///////////////////////////////////////////////////////
	-- FSM3 SECTION
	-- Round Finalization / Load new block / Output generation		 
	
	fsm3_output : entity work.ppl_output_unit(struct) 
		generic map ( h => HS, w => w, ppl => ppl ) 
		port map ( clk => clk, rst => rst,
					hash_result => rdprime, id_in => id_in,
					sid => sid, shift_dout => shift_dout, load_hash => load_hash, 
					load_id => load_id, sout => sout,
					dout => dout );			 
					
					
	id_in_gen : for i in 0 to ppl-1 generate
		id_in((ppl-i)*id_size-1 downto (ppl-i-1)*id_size) <= din(ppl*w-(i+1)*w+15 downto ppl*w-(i+1)*w);
	end generate; 		
								 
end struct;