-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

-- Possible generic(s) values: 												 
-- 		FF = {1, 2} {FF = 2 by default}
--		PPL = {2, 4}
--		HS = {HASH_SIZE_256, HASH_SIZE_512}

-- Valid combinations :
--		(FF, PPL) = (2, 2), (2, 4), (1,2), (1,4)

-- Extra generic(s) :
-- 		w = {2^x} where x can be any reasonable number. By default, x is 6
-- Note : Input and output test vectors must correspond to the size of w

library ieee;
use ieee.std_logic_1164.all; 
use work.sha3_pkg.all;
use work.sha3_blake_package.all;

entity blake_top_ppl is		
	generic (			   
		FF : integer := 2;		
		PPL : integer := 2;
		HS : integer := 256;
		W : integer := 64
	); 
	port (		
		-- global
		rst 	: in std_logic;
		clk 	: in std_logic;	
		--fifo
		src_ready : in std_logic_vector(ppl-1 downto 0);
		src_read  : out std_logic_vector(ppl-1 downto 0);
		dst_ready : in std_logic;
		dst_write : out std_logic;		
		din		: in std_logic_vector(ppl*w-1 downto 0);
		dout	: out std_logic_vector(w-1 downto 0)
	);	   
end blake_top_ppl;

architecture structure of blake_top_ppl is  
	constant BS : integer := get_b( HS );
	constant IW : integer := get_iw( HS );	
	constant version : integer := SHA3_ROUND3;
	-- fsm1						   
	signal extra_block : std_logic_vector(ppl-1 downto 0);
	signal ein, lc, ec, lm: std_logic_vector(ppl-1 downto 0);
	signal dt, eth, etl, final_segment : std_logic_vector(ppl-1 downto 0);
	signal zcRemainder, zc0, zc1, stbot : std_logic_vector(ppl-1 downto 0);
	-- fsm2
	signal load_h : std_logic;
	signal round : std_logic_vector(3+FF-1 downto 0); 	 
	signal sf, sperm : std_logic;
	signal eh : std_logic_vector(ppl-1 downto 0);
	signal smsg : std_logic_vector(log2(ppl)-1 downto 0);
	-- fsm3
	signal sout : std_logic_vector(log2(ppl)-1 downto 0);	
	signal sid : std_logic;
	signal shift_dout : std_logic_vector(ppl-1 downto 0);
	signal load_id, load_hash : std_logic_vector(ppl-1 downto 0);
									 
begin	
	fseg : for i in 0 to ppl-1 generate
		final_segment(i) <= din(w*ppl-i*w-1);	-- MSB of a word
	end generate; 
	
	control_gen : entity work.blake_control_ppl(struct) 
		generic map ( VERSION => VERSION, IW => IW, BS => BS, HS => HS, W => W, FF => FF, PPL => PPL )
		port map (
		rst => rst, clk => clk, 
		src_ready => src_ready, src_read => src_read, dst_ready => dst_ready, dst_write => dst_write,	 
		final_segment => final_segment, zcRemainder => zcRemainder, zc0 => zc0, zc1 => zc1, stbot=> stbot, ein => ein, ec => ec, 
			lc => lc, dt => dt, eth => eth, etl => etl, extra_block => extra_block, lm => lm,
		load_h => load_h, eh => eh, smsg => smsg, sf => sf, round => round, sperm => sperm,
		shift_dout => shift_dout, load_id => load_id, load_hash => load_hash,  sout => sout,  sid => sid
	);			
	
	datapath_gen : entity work.blake_datapath_ppl(struct) 	
		generic map ( BS => BS, IW => IW, HS => HS, w => w, PPL => PPL, FF => FF )
		port map (
		rst => rst, clk => clk, din => din, dout => dout, 
		zcRemainder => zcRemainder, zc0 => zc0, zc1 => zc1, stbot=> stbot, ein => ein, ec => ec, 
			lc => lc, dt => dt, eth => eth, etl => etl, extra_block => extra_block, lm => lm,
		load_h => load_h, eh => eh, smsg => smsg, sf => sf, round => round,	sperm => sperm,
		shift_dout => shift_dout, load_id => load_id, load_hash => load_hash,  sout => sout,  sid => sid
	);

end structure;
	
	