-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

-- Version number 0.1

library ieee;
use ieee.std_logic_1164.ALL;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.ALL;
use ieee.std_logic_textio.all;
use work.sha3_pkg.all;
use work.sha_tb_pkg_ppl.all;

LIBRARY std;
use std.textio.all;				 
					
ENTITY sha_tb_all_ppl IS		
END sha_tb_all_ppl;

ARCHITECTURE behavior OF sha_tb_all_ppl IS 

	type data_type is array ( 0 to ppl-1 ) of std_logic_vector(w-1 downto 0);
	type verification_mode is ( check_id, check_msg );
	-- ==================== --
	-- Verification section --
	-- ==================== --
	type verification_ram_type is array ( 0 to max_msg-1 ) of std_logic_vector(hashsize-1 downto 0);
	type verification_valid_type is array ( 0 to max_msg-1 ) of std_logic;
	signal vram : verification_ram_type;	
	function init_valid return verification_valid_type is
		variable ret : verification_valid_type;
	begin
		for i in 0 to max_msg-1 loop
			ret(i) := '0';
		end loop;
		return ret;
	end function init_valid;
	signal vram_valid, vram_checked : verification_valid_type := init_valid;
	-- =================== --
	-- STR DECLARATION --
	-- =================== --
	constant pad_str 		: string := get_pad_str( pad_mode );
	constant split_str 		: string := get_split_str( split_type );
	constant interface_str 	: string := get_interface_str( interface_type );					
	constant ifile_name 	: string := split_str & "_" & pad_str & "_" & algorithm & "_PPL_datain_h" & integer'image(hashsize) & "_w" & integer'image(w) & ".txt";
	constant ofile_name 	: string := split_str & "_" & pad_str & "_" & algorithm & "_PPL_dataout_h" & integer'image(hashsize) & "_w" & integer'image(w) & ".txt";

	-- =================== --
	-- SIGNALS DECLARATION --
	-- =================== --
	
	-- simulation signals (used by ATHENa script, ignore if not used)
	signal simulation_pass : std_logic := '0'; 	  	-- '0' signifies a pass at the end of simulation, '1' is fail
	signal stop_clock : boolean := false;		-- '1' signifies a completed simulation, '0' otherwise
	signal force_exit : boolean := false;
	
	-- error check signal
	signal global_stop : std_logic := '1';
	-- globals
	SIGNAL clk :  std_logic := '0';	  
	signal io_clk : std_logic := '0';
	SIGNAL rst :  std_logic := '0';
	
	--Inputs
	SIGNAL fifoin_write : std_logic_vector(ppl-1 downto 0) := (others => '0');
	signal fifoout_read : std_logic := '0';
	SIGNAL ext_idata :  std_logic_vector(w-1 downto 0);
	
	--Outputs
	SIGNAL fifoout_empty :  std_logic;
	SIGNAL fifoin_full :  std_logic_vector(ppl-1 downto 0);
	SIGNAL ext_odata :  std_logic_vector(w-1 downto 0);
	
	-- Internals
	SIGNAL fifoin_empty, fifoin_empty_hold :  std_logic_vector(ppl-1 downto 0);
	SIGNAL fifoin_read 	:  std_logic_vector(ppl-1 downto 0);
	SIGNAL fifoout_full :  std_logic;
	SIGNAL fifoout_write :  std_logic;	
	SIGNAL odata 		: std_logic_vector(w-1 downto 0);	   
	SIGNAL idata 		:  data_type;
	
	SIGNAL fifoin_write_sig, fifoin_full_sig : std_logic;
	SIGNAL msg_select : std_logic_vector(log2(ppl)-1 downto 0) := (others =>'0');
	
	------------- clock constant ------------------
	constant io_clk_period : time := get_io_clk( clk_period, fr, interface_type );
	----------- end of clock constant -------------	  
	
	------------- string constant ------------------
   	constant cons_len 	: string(1 to 6) := "Len = ";
	constant cons_msg 	: string(1 to 6) := "Msg = ";		   
	constant cons_mid 	: string(1 to 6) := "MID = ";
	constant cons_md 	: string(1 to 5) := "MD = ";	
	constant cons_id 	: string(1 to 5) := "ID = ";
	constant cons_eof 	: string(1 to 5) := "#EOF#";
	constant cons_blank	: string(1 to 5) := "     ";
	----------- end of string constant -------------		 

	------------- debug constant ------------------
   	constant debug_input : boolean := false;
	constant debug_output : boolean := false;
	----------- end of clock constant -------------

	-- ================= --
	-- FILES DECLARATION -- 
	-- ================= --
    
	--------------- input / output files -------------------
	FILE datain_file	: TEXT OPEN READ_MODE   is  ifile_name; 
	FILE dataout_file	: TEXT OPEN READ_MODE   is  ofile_name; 
	
	FILE log_file : TEXT OPEN WRITE_MODE is logfile_name;  
	FILE result_file : TEXT OPEN WRITE_MODE is resultfile_name;
	------------- end of input files --------------------
	signal hold_data : std_logic := '1';
	
BEGIN
	
	clk_gen: process  
	begin
		if ((not stop_clock) and (global_stop = '1')) then 
			clk <= '1'; 
			wait for clk_period/2; 
			clk <= '0'; 
			wait for clk_period/2; 
		else 
			wait; 
		end if; 
	end process clk_gen; 

	io_clk_gen: process  
	begin
		if ((not stop_clock) and (global_stop = '1')) then 
			io_clk <= '1'; 
			wait for io_clk_period/2; 
			io_clk <= '0'; 
			wait for io_clk_period/2; 
		else 
			wait; 
		end if; 
	end process io_clk_gen; 
	
	-- ============ --
	-- PORT MAPPING --
	-- ============ --		 		
	
	sha3unit:  entity work.wrapper(struct)
		generic map ( fifodepth => depth, hashsize => hashsize, w => w, ppl => ppl )
		port map (		
			-- global signals
			rst => rst,
			clk => clk,
			-- input fifo signals
			ext_idata => ext_idata,
			fifoin_write => fifoin_write,
			fifoin_full => fifoin_full,
			-- output fifo signals
			fifoout_read => fifoout_read,
			ext_odata => ext_odata,
			fifoout_empty => fifoout_empty,		   
			-- simulation 
			hold_data => hold_data
		);
	
	 -- =================== --
	 -- END OF PORT MAPPING --
	 -- =================== --

    -- ===========================================================
	-- ==================== DATA POPULATION ====================== 
	tb_writedata : PROCESS
   		VARIABLE 	line_data, errorMsg	: 	LINE;
		variable 	word_block 				:  	std_logic_vector(w-1 downto 0) := (others=>'0');
		variable 	read_result				: 	boolean; 
		variable	loop_enable				: 	std_logic := '1';	   
		variable	temp_read				: 	string(1 to 6); 
		variable	valid_line				: 	boolean := true;
	BEGIN	
		rst <= '1';	   		wait for 5*clk_period;
		rst <= '0';	   		wait for clk_period;	
		
		-- read header
		while ( not endfile (datain_file)) and ( loop_enable = '1' ) loop
			if endfile (datain_file) then
				loop_enable := '0';
			end if;
			
			readline(datain_file, line_data);
			read(line_data, temp_read, read_result);	 
			if (temp_read = cons_mid) then
				loop_enable := '0';
			end if;
		end loop;
		
		-- do operations in the falling edge of the io_clk
		wait for io_clk_period/2; 
		
		while not endfile ( datain_file ) loop  		    
			-- if the fifo is full, wait ...	
			fifoin_write_sig <= '1';	
			if ( fifoin_full_sig = '1' ) then		
				fifoin_write_sig <= '0';			
				wait until fifoin_full_sig <= '0';	 
				wait for io_clk_period/2; -- write in the rising edge
				fifoin_write_sig <= '1';			
			end if;	  	
			
			hread( line_data, word_block, read_result );
			while (((read_result = false) or (valid_line = false)) and (not endfile( datain_file ))) loop
				readline(datain_file, line_data);
				read(line_data, temp_read, read_result);	-- read line header 
				if ( temp_read /= cons_msg and temp_read /= cons_len and temp_read /= cons_mid) then
					valid_line := false;	   	  
					fifoin_write_sig <= '0';	
				else
					if temp_read = cons_mid then
						if msg_select = ppl-1 then
							msg_select <= (others => '0');
						else
							msg_select <= msg_select + 1;
						end if;						
					end if;
					valid_line := true;		 
					fifoin_write_sig <= '1';	
				end if;
				hread( line_data, word_block, read_result );-- read data 		
				report "---------din:reading new line--------" severity error;
			end loop;			
			ext_idata <= word_block;
	   		wait for io_clk_period; 
		end loop;   	   		
		fifoin_write_sig <= '0';
		wait for 10*io_clk_period; -- wait another 10 clock cycles before start processing		
		wait;
	END PROCESS;
	
	hold_proc : process
	begin		
		report "---------Loading Data-----------" severity error;
		wait for 1000*io_clk_period;
		hold_data <= '0';									  
		report "------Loading Data Finished-----" severity error;
		wait;
	end process;
	
	process( msg_select, fifoin_write_sig, fifoin_full )
	begin					
		fifoin_write <= (others => '0');
		fifoin_write(conv_integer(msg_select)) <= fifoin_write_sig;
		fifoin_full_sig <= fifoin_full(conv_integer(msg_select));
	end process;
	
	
	-- ===========================================================		
	
	
	-- ===========================================================		
	-- =================== DATA VERIFICATION =====================
	populate_vram : PROCESS	  
		variable 	line_data	: 	LINE;
		variable 	hashvalue 			:  	std_logic_vector(hashsize-1 downto 0) := (others=>'0');
		variable 	idvalue 			:  	std_logic_vector(w-1 downto 0) := (others=>'0');
		variable	temp_read			: 	string(1 to 5); 	
		variable 	read_result			: 	boolean; 
		variable 	vram_addr		: integer;
	begin	
		vram_addr := 1;
		while ( not endfile (dataout_file)) and (not force_exit) loop
			readline(dataout_file, line_data);
			read(line_data, temp_read);	 
			while( temp_read /= cons_md and temp_read /= cons_eof and temp_read /= cons_id ) loop
				readline(dataout_file, line_data);
				read(line_data, temp_read);	 
			end loop;				 
			if (temp_read = cons_id) then
				hread( line_data, idvalue, read_result );
				vram_addr := conv_integer(idvalue);
			elsif (temp_read = cons_md) then
				hread( line_data, hashvalue, read_result );
				vram(vram_addr) <= hashvalue;
				vram_valid(vram_addr) <= '1';	
				vram_addr := vram_addr + 1;				
			elsif ( temp_read = cons_eof ) then
				force_exit <= true;					
			end if;
			temp_read := cons_blank;
		end loop;
		wait;
	end process;  	

	tb_verifydata : process
		variable errorMsg	: 	LINE;
		variable finished : std_logic := '0';		
		variable mode : verification_mode;
		variable vram_addr : integer;
		variable msg_id, expected_hash_word : std_logic_vector(w-1 downto 0);
		variable parity_cnt, msg_fail : integer;
		variable FAIL : integer := 0;
	begin		
		wait for 10*io_clk_period;
		mode := check_id;
		-- init checked 
		for i in 0 to max_msg-1 loop
			vram_checked(i) <= '0';
		end loop;
		while ( finished = '0') loop
			if mode = check_id then
				wait until ( falling_edge( clk ) and fifoout_empty = '0' );	   

				fifoout_read <= '1';
				msg_id := ext_odata;
				vram_addr := conv_integer(msg_id(15 downto 0));
				
				-- Check for valid msg_id
				parity_cnt := 0;				
				for i in 16 to w-1 loop
					if ( msg_id(i) = '1' ) then
						parity_cnt := parity_cnt + 1;
					end if;
				end loop;				
				if ( parity_cnt > 2**16 ) then
					finished := '1'; FAIL := 1;
					report "FAIL (1): The output message id is bigger than 2^16. I doubt you have this message in the testbench." severity error;
				else		
					-- Check for validity				
					if ( vram_valid(conv_integer(msg_id(15 downto 0))) = '0' ) then
						finished := '1'; FAIL := 1; 						
						report "FAIL (1): Invalid ID => " & integer'Image(conv_integer(msg_id(5 downto 0))) & ". This message ID is not part of the testbench." severity error;
						stop_clock <= true;	
					elsif ( vram_checked(conv_integer(msg_id(15 downto 0))) = '1' ) then
						finished := '1'; FAIL := 1; stop_clock <= true;
						report "FAIL (1): Invalid ID => " & integer'Image(conv_integer(msg_id(5 downto 0))) & ". This message ID has already been checked."   severity error;
					else
						mode := check_msg;
						wait for io_clk_period;
					end if;	
				end if;
				fifoout_read <= '0';
				-- set flag
				vram_checked(vram_addr) <= '1';
			elsif mode = check_msg then				
				wait until ( falling_edge( clk ) and fifoout_empty = '0' );
				
				-- report "Just checking out the time" severity failure; 
				msg_fail := 0;
				for i in 0 to hashsize/w-1 loop					
					expected_hash_word := vram(vram_addr)(w*(hashsize/w)-i*w-1 downto w*(hashsize/w-1)-i*w);
					fifoout_read <= '1';										
					wait for 1/2*io_clk_period;					
					if ( expected_hash_word /= ext_odata ) then
						finished := '1'; FAIL := 1; msg_fail := 1;
						report "FAIL (1): Message ID (" & integer'Image(vram_addr) & ") contains invalid hash data in word number --> " & integer'Image(i) severity error;
						--write(errorMsg, string'("Expected valued => "));
--						hwrite(errorMsg, vram(conv_integer(msg_id))(w*(hashsize/w)-1 downto w*(hashsize/w-1)) );
--						write(errorMsg, string'(" :: Received valued => "));
--						hwrite(errorMsg, ext_odata );
--						report line'image(errorMsg) severity error;
						stop_clock <= true;	
					end if;
					
					if (fifoout_empty = '1') then
						fifoout_read <= '0';
						wait until ( falling_edge( clk ) and fifoout_empty = '0' );
					else 
						wait for io_clk_period;
					end if;			
					
				end loop;	
				if ( msg_fail = 0 ) then
					report "Hooray!! The following message ID has passed the test =>" & integer'Image(vram_addr) severity error;
				else
					report "Ouch!! You'll want to debug Message ID (" & integer'Image(vram_addr) & ") as it fails this testbench" severity error;
				end if;
				-- check for finish
				finished := '1';
				for i in 0 to max_msg-1 loop					
					if ( vram_valid(i) /= vram_checked(i) ) then
						finished := '0';						
					end if;
				end loop;								
				mode := check_id;
			end if;
		end loop;			   
		if ( FAIL = 0 ) then
			report "PASS (0): Hooray!! You have successfully pass the test." severity error;
			report "PASS (0): SIMULATION FINISHED || Input/Output files :: " & ifile_name & "/" & ofile_name severity error;
		else
			report "FAIL (1): Simulation failed to pass the testbench || Input/Output files :: " & ifile_name & "/" & ofile_name severity error;
		end if;
		stop_clock <= true;
		wait;
	end process;

		  
	errcheck : process
		variable pad_pass, split_pass, algo_pass, interface_pass, fifo_pass : integer := 0;
	begin
		if (( pad_mode /= pad ) and ( pad_mode /= nopad )) then
			pad_pass := 1;
		end if;							  
		if (( split_type /= no_split ) and ( split_type /= fixed_split ) and ( split_type /= random_split )) then
			split_pass := 1;
		end if;							  																			   
		if ((interface_type /= one_clk) and (interface_type /= two_clk)) then
			interface_pass := 1;
		end if;	
		if ((fifo_mode > 2) or (fifo_style > 1)) then
			fifo_pass := 1;
		end if;
		
		if ( pad_pass = 1 or split_pass = 1 or algo_pass = 1 or interface_pass = 1  or fifo_pass = 1 ) then
			if ( pad_pass = 1 ) then 
				report "Invalid padding mode ==> "& pad_str &". Only pad or nopad mode is allowed." severity warning;			
			end if;
			if ( split_pass = 1 ) then
				report "Invalid split type ==> " & split_str & ". Only no_split, fixed_split or random_split is allowed." severity warning;		
			end if;
			if ( interface_pass = 1 ) then
				report	"Invalid interface type ==> " & interface_str & ". Valid interface types are one_clk or two_clk" severity warning;
			end if;
			if ( fifo_pass = 1 ) then
				report "Invalid fifo mode or fifo style" severity warning;
			end if;			
			
			report "Invalid core parameters!! Please see above for the list." severity failure;
			global_stop <= '0';
		end if;
		wait;
	end process;
-- ================================================================	
-- # This section is used for testing the tb without the core.  
-- # Must comment below process in order to run the core properly.
-- # Valid signals that can be manipulated in the below process are the handshake signals from a hash core.
-- # They are -- fifoin_read, fifoout_empty			   
-- ================================================================	
--	inputtest : process
--	begin	
--		fifoin_read <= '1';
--		fifoout_empty <= '0';
--		wait;
--	end process;
END;
